#!/usr/bin/env python
# coding: utf-8

# Some L-systems:
#   https://www.robertdickau.com/lsys2d.html
#   https://resources.wolframcloud.com/FunctionRepository/resources/LSystemPlot/

# In[12]:
import math

def get_l_system_paths(depth, theta, axiom, ignore, rules, dtheta,
        initial_theta, length_scale):
    '''Apply an L-system.'''
    # Build the final script.
    script = axiom
    for i in range(depth):
        new_script = ''
        for ch in script:
            if ch in rules:
                new_script += rules[ch]
            else:
                new_script += ch
        script = new_script
    # print(script)

    # Create the points.
    p1 = (0, 0)
    length = 100
    paths = []      # A list of paths
    path = [p1]     # The current path
    theta = initial_theta
    waypoints = []  # Places we will go back to
    wayangles = []  # Angles for the waypoints
    waylengths = []
    for ch in script:
        if ch in ignore:
            pass
        elif ch == '-':
            theta -= dtheta
        elif ch == '+':
            theta += dtheta
        elif ch == '[':
            waypoints.append(path[-1])
            wayangles.append(theta)
            waylengths.append(length)
        elif ch == ']':
            if len(path) > 1: paths.append(path)
            p1 = waypoints.pop()
            path = [p1]
            theta = wayangles.pop()
            length = waylengths.pop()
        else:
            x = p1[0] + length * math.cos(theta)
            y = p1[1] + length * math.sin(theta)
            p1 = (x, y)
            path.append(p1)
            length *=  length_scale

    if len(path) > 1:
        paths.append(path)

    # Return the points.
    return paths

#%%

def draw_point(canvas, pt, radius, outline, fill):
    canvas.create_oval(
        pt[0] - radius,
        pt[1] - radius,
        pt[0] + radius,
        pt[1] + radius,
        outline=outline,
        fill=fill)

def fit_curves_to_canvas(canvas, curves, margin):
    '''Scale a set of curves to fit a canvas.'''
    # Get the canvas's dimensions.
    canvas_wid = canvas.winfo_width() - 2 * margin
    canvas_hgt = canvas.winfo_height() - 2 * margin

    # Get the curves' dimensions.
    x1, y1, x2, y2 = curves_bbox(canvas, curves)
    curve_wid = x2 - x1
    curve_hgt = y2 - y1

    # Calculate the scale factor.
    x_scale = (canvas_wid - 2 * margin) / curve_wid
    y_scale = (canvas_hgt - 2 * margin) / curve_hgt
    scale = min(x_scale, y_scale)

    # Scale the curve.
    for curve in curves:
        canvas.scale(curve, 0, 0, scale, scale)

    # Center the curves.
    x1, y1, x2, y2 = curves_bbox(canvas, curves)
    cx = (x1 + x2) / 2
    cy = (y1 + y2) / 2
    xmin = cx - canvas_wid / 2
    xmax = cx + canvas_wid / 2
    ymin = cy - canvas_hgt / 2
    ymax = cy + canvas_hgt / 2
    canvas.config(scrollregion=(xmin, ymin, xmax, ymax))

def curves_bbox(canvas, curves):
    '''Return a bounding box for a set of curves.'''
    x1, y1, x2, y2 = canvas.bbox(curves[0])
    for curve in curves:
        new_x1, new_y1, new_x2, new_y2 = canvas.bbox(curve)
        if new_x1 < x1: x1 = new_x1
        if new_x2 > x2: x2 = new_x2
        if new_y1 < y1: y1 = new_y1
        if new_y2 > y2: y2 = new_y2
    return x1, y1, x2, y2

# In[18]:
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox

class LSystemsApp:
    def __init__(self):
        self.window = tk.Tk()
        self.window.title('l_systems_plus')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('400x400')

        # Build the UI.
        frame = tk.Frame(self.window)
        frame.pack(side=tk.TOP, anchor=tk.W, padx=3, pady=3)

        label = tk.Label(frame, text='Depth:')
        label.pack(side=tk.LEFT, padx=3)
        self.depth_var = tk.StringVar(value='3')
        entry = tk.Entry(frame, textvariable=self.depth_var, width=3)
        entry.pack(side=tk.LEFT, padx=3)

        label = tk.Label(frame, text='Curve:')
        label.pack(side=tk.LEFT, padx=3)

        # Each L-system includes:
        #   Name: value
        # Where value includes:
        #   Dictionary of rules
        #   Axiom
        #   Ignore
        #   Dtheta
        #   Max Depth
        #   Theta 0
        #   Length Scale
        self.lsystems = {
            'Tree': (
                {'F': 'F[-F][+F]'},
                'F', '', 45, 10, -90, 0.75),
            # https://en.wikipedia.org/wiki/L-system#Example_7:_fractal_plant
            # The original version uses axiom X and angle 25. I've modified
            # it to give it a proper orientation in our coordinate system.
            'Fern': (
                {'X': 'F+[[X]-X]-F[-FX]+X', 'F': 'FF'},
                'X', 'X', -25, 9, -50, 1),
            '32-Segment Curve': (
                {'F': '-F+F-F-F+F+FF-F+F+FF+F-F-FF+FF-FF+F+F-FF-F-F+FF-F-F+F+F-F+'},
                'F+F+F+F', '', 90, 4, 0, 1),
            'Box Fractal': (
                {'F': 'F-F+F+F-F'},
                'F-F-F-F', '', 90, 7, 0, 1),
            'Dragon Curve': (
                {'X': 'X+YF+', 'Y': '-FX-Y'},
                'FX', '', 90, 20, 0, 1),
            # Hilbert Curve: https://www.fractal.garden/l-system/hilbert-curve
            'Hilbert Curve': (
                {'V': '-WF+VFV+FW-', 'W': '+VF-WFW-FV+'},
                'W', 'VW', 90, 10, 0, 1),
            'Hilbert Curve II': (
                {'X': 'XFYFX+F+YFXFY-F-XFYFX', 'Y': 'YFXFY-F-XFYFX+F+YFXFY'},
                'X', '', 90, 6, 0, 1),
            'Koch Snowflake': (
                {'F': 'F+F--F+F'},
                'F--F--F', '', 60, 7, 0, 1),
            'Peano Curve': (
                {'F': 'F+F-F-F-F+F+F+F-F'},
                'F', '', 90, 6, 0, 1),
            'Peano-Gosper Curve': (
                {'X': 'X+YF++YF-FX--FXFX-YF+', 'Y': '-FX+YFYF++YF+FX--FX-Y'},
                'FX', '', 60, 7, 0, 1),
            'Gosper Boundary': (
                {'A': 'A-A+A'},
                'A-A-A-A-A-A-', '', 60, 7, 0, 1),
            'Quadratic Koch Island': (
                {'F': 'F-F+F+FFF-F-F+F'},
                'F+F+F+F', '', 90, 6, 0, 1),
            'Sierpiński Arrowhead': (
                {'X': 'YF+XF+Y', 'Y': 'XF-YF-X'},
                'YF', '', 60, 10, 0, 1),
            # Sierpiński Curve: https://en.wikipedia.org/wiki/Sierpi%C5%84ski_curve
            'Sierpiński Curve': (
                {'X': 'XF+G+XF--F--XF+G+X'},
                'F--XF--F--XF', 'X', 45, 9, 0, 1),
            # Sierpiński Square Curve: https://en.wikipedia.org/wiki/Sierpi%C5%84ski_curve
            'Sierpiński Square Curve': (
                {'X': 'XF-F+F-XF+F+XF-F+F-X'},
                'F+XF+F+XF', '', 90, 9, 0, 1),
            'Siepiński Sieve': (
                {'F': 'FF', 'X': '--FXF++FXF++FXF--'},
                'FXF--FF--FF', '', 60, 10, 0, 1),
        }

        curve_names = list(self.lsystems.keys())
        self.curve_combo = ttk.Combobox(frame, values=curve_names,
                                        state='readonly')
        self.curve_combo.pack(side=tk.LEFT, padx=3)
        self.curve_combo.current(0)

        button = tk.Button(frame, text='Draw', width=10, command=self.draw)
        button.pack(side=tk.LEFT, padx=10)

        self.canvas = tk.Canvas(self.window, bg='white',
            borderwidth=2, relief=tk.SUNKEN)
        self.canvas.pack(padx=10, pady=(0, 10),
            side=tk.BOTTOM, fill=tk.BOTH, expand=True)

        # If the user presses Enter, draw.
        entry.bind('<Return>', lambda event: self.draw())

        # Display the window.
        entry.focus_force()
        self.window.mainloop()

    def kill_callback(self):
        # Destroy the tkinter window.
        self.window.destroy()

    def draw(self):
        # Delete any previous drawing objects.
        self.canvas.delete(tk.ALL)

        # Get the desired depth.
        try:
            depth = int(self.depth_var.get())
        except:
            depth = -1

        # See which curve is selected.
        curve_name = self.curve_combo.get()
        lsystem = self.lsystems[curve_name]
        rules, axiom, ignore, dtheta, max_depth, \
            initial_theta, length_scale = lsystem
        dtheta = math.radians(dtheta)
        theta = math.radians(initial_theta)

        # See if it's the Sierpiński Arrowhead.
        if curve_name == 'Sierpiński Arrowhead':
            # Draw odd-numbered depths backward so they all point down.
            if depth % 2 == 1:
                theta = math.radians(180)

        # Validate the desired depth.
        if depth < 0 or depth > max_depth:
            messagebox.showinfo('Depth Error',
                f'{curve_name} depth must be an integer between 0 and {max_depth}.')
            return

        # Generate the curve.
        paths = get_l_system_paths(depth, theta, axiom, ignore, rules, \
                                   dtheta, theta, length_scale)

        # Draw the path(s).
        num_points = 0
        curves = []
        for path in paths:
            # print(path)
            curves.append(self.canvas.create_line(path, width=0, fill='black'))
            num_points += len(path)

        # Scale to fit.
        fit_curves_to_canvas(self.canvas, curves, 10)

        # Print the number of pathcs and points.
        print(f'{len(paths):,} paths, {num_points:,} points')

# In[19]:
LSystemsApp()
print('Done')
